#include <postgres.h>
#include <fmgr.h>
#include <access/skey.h>
#include <utils/builtins.h>
#include <limits.h>

PG_MODULE_MAGIC;

static Datum entries[1];

typedef struct TypeInfo {
    bool    is_varlena;
	Datum   (*leftmostvalue)();
	Datum   (*typecmp)(FunctionCallInfo);
} TypeInfo;

typedef struct QueryInfo {
	StrategyNumber		strategy;
	Datum				datum;
} QueryInfo;

#define  GIN_EXTRACT_VALUE(type)												\
PG_FUNCTION_INFO_V1(gin_extract_value_##type);									\
Datum       gin_extract_value_##type(PG_FUNCTION_ARGS);							\
Datum																			\
gin_extract_value_##type(PG_FUNCTION_ARGS)										\
{																				\
	Datum datum = PG_GETARG_DATUM(0);											\
	int32 *nentries = (int32 *) PG_GETARG_POINTER(1);							\
																				\
	if ( TypeInfo_##type.is_varlena )											\
		datum = PointerGetDatum(PG_DETOAST_DATUM(datum));						\
	entries[0] = datum;															\
																				\
	*nentries = 1;																\
																				\
	PG_RETURN_POINTER(entries);													\
}

#define GIN_EXTRACT_QUERY(type)													\
PG_FUNCTION_INFO_V1(gin_extract_query_##type);									\
Datum       gin_extract_query_##type(PG_FUNCTION_ARGS);							\
Datum																			\
gin_extract_query_##type(PG_FUNCTION_ARGS)										\
{																				\
	Datum datum = PG_GETARG_DATUM(0);											\
	int32 *nentries = (int32 *) PG_GETARG_POINTER(1);							\
	StrategyNumber strategy = PG_GETARG_UINT16(2);								\
	bool	**partialmatch = (bool**) PG_GETARG_POINTER(3);						\
	bool	*ptr_partialmatch;													\
	void	***extra_data= (void***) PG_GETARG_POINTER(4);						\
	QueryInfo	*data = (QueryInfo*)palloc(sizeof(QueryInfo));					\
																				\
	*nentries = 1;																\
	ptr_partialmatch = *partialmatch = palloc(sizeof(bool));					\
	*ptr_partialmatch = false;													\
	if ( TypeInfo_##type.is_varlena )											\
			datum = PointerGetDatum(PG_DETOAST_DATUM(datum));					\
	data->strategy = strategy;													\
	data->datum = datum;														\
	*extra_data = (void***) palloc(sizeof(void**));								\
	**extra_data = data;														\
																				\
	switch(strategy) {															\
		case	BTLessStrategyNumber:											\
		case	BTLessEqualStrategyNumber:										\
			entries[0] = TypeInfo_##type.leftmostvalue();						\
			*ptr_partialmatch = true;											\
			break;																\
		case	BTGreaterEqualStrategyNumber:									\
		case	BTGreaterStrategyNumber:										\
			*ptr_partialmatch = true;											\
		case	BTEqualStrategyNumber:											\
			entries[0] = datum;													\
			break;																\
		default:																\
			elog(ERROR,"1Unknown strategy number, %d", strategy);				\
	}																			\
																				\
	PG_RETURN_POINTER(entries);													\
}
			
#define GIN_COMPARE_PREFIX(type)												\
PG_FUNCTION_INFO_V1(gin_compare_prefix_##type);									\
Datum       gin_compare_prefix_##type(PG_FUNCTION_ARGS);						\
Datum																			\
gin_compare_prefix_##type(PG_FUNCTION_ARGS)										\
{																				\
	Datum 	a = PG_GETARG_DATUM(0);												\
	Datum 	b = PG_GETARG_DATUM(1);												\
	QueryInfo *data = (QueryInfo*)PG_GETARG_POINTER(2);							\
	int32	res;																\
																				\
	res = DatumGetInt32(DirectFunctionCall2(									\
				TypeInfo_##type.typecmp, 										\
				(data->strategy == BTLessStrategyNumber || 						\
					data->strategy == BTLessEqualStrategyNumber )				\
						? PointerGetDatum(data->datum) : a, 					\
				b																\
		));																		\
																				\
	switch(data->strategy) {													\
		case	BTLessStrategyNumber:											\
			if ( res > 0 )														\
				res = 0;														\
			else																\
				res = 1;														\
			break;																\
		case	BTLessEqualStrategyNumber:										\
			if ( res >= 0 )														\
				res = 0;														\
			else																\
				res = 1;														\
			break;																\
		case	BTEqualStrategyNumber:											\
			if ( res !=0  )														\
				res = 1;														\
			break;																\
		case	BTGreaterEqualStrategyNumber:									\
			if ( res <= 0 )														\
				res = 0;														\
			else																\
				res = 1;														\
			break;																\
		case	BTGreaterStrategyNumber:										\
			if ( res < 0 )														\
				res = 0;														\
			else if ( res == 0 )												\
				res = -1;														\
			else																\
				res = 1;														\
			break;																\
		default:																\
			elog(ERROR,"2Unknown strategy number, %d", data->strategy);			\
	}																			\
																				\
	PG_RETURN_INT32(res);														\
}

#define GIN_SUPPORT(type)			\
	GIN_EXTRACT_VALUE(type)			\
	GIN_EXTRACT_QUERY(type)			\
	GIN_COMPARE_PREFIX(type)


PG_FUNCTION_INFO_V1(gin_btree_consistent);
Datum       gin_btree_consistent(PG_FUNCTION_ARGS);
Datum
gin_btree_consistent(PG_FUNCTION_ARGS)
{
	bool       *recheck = (bool *) PG_GETARG_POINTER(3);
	*recheck = false;
	PG_RETURN_BOOL(true);
}

static Datum 
leftmostvalue_int4()
{
	return Int32GetDatum(INT_MIN);
}
static TypeInfo TypeInfo_int4 = {false, leftmostvalue_int4, btint4cmp};
GIN_SUPPORT(int4)

static Datum
leftmostvalue_text()
{
	return PointerGetDatum(cstring_to_text_with_len("", 0));
}
static TypeInfo TypeInfo_text = {true, leftmostvalue_text, bttextcmp};
GIN_SUPPORT(text)
